
from .request_rotb_compute_strategy_base import RequestRotbComputeStrategyBase
from .response_rotb_compute_strategy_prices import ResponseRotbComputeStrategyPrices

LIMIT_SPLIT_REQUEST = 500


class RequestRotbComputeStrategyPrices(RequestRotbComputeStrategyBase):
    """
    """

    def __init__(self,
                 dic=None,
                 date_fallback_next=True,
                 limit_split=LIMIT_SPLIT_REQUEST):
        """
        """
        dic_url = {'base_url': 'https://analytics-api.sgmarkets.com',
                   'service': '/rotb',
                   'endpoint': '/v1/compute-strategy/prices'}
        super().__init__(dic,
                         date_fallback_next,
                         limit_split,
                         dic_url,
                         ResponseRotbComputeStrategyPrices)




import os
import json
import copy

import numpy as np
import datetime as dt
import pandas as pd

from ._util import Util

from IPython.display import Markdown
from sgmarkets_api_auth.util import save_result


class ResponseRotbComputeStrategyComponents:
    """
    """

    def __init__(self,
                 li_raw_data=None,
                 obj_req=None):
        """
        """
        assert isinstance(li_raw_data, list), \
            'Error: li_raw_data must be a list - Run call_api() again with debug=True'
        for dic_res in li_raw_data:
            assert isinstance(dic_res, dict), \
                'Error: Each dic_res must be a list - Run call_api() again with debug=True'
            assert 'componentSeries' in dic_res, \
                'Error: componentSeries must be a key of each dic_res - Run call_api() again with debug=True'

        raw_data = []
        # raw_data_analysis=[]
        for dic_res in li_raw_data:
            raw_data += dic_res['componentSeries']

        self.raw_data = copy.deepcopy(raw_data)
        self.obj_req = obj_req

        self.df_req, self.df_res, self.df_set, self.df_req_ra, self.df_res_ra, self.df_set_ra = self._build_df_res_req()
        self.dic_req_param, self.dic_res_param = self._build_dic_param()

    def _get_dates(self, df_res):
        """
        """
        dic = self.obj_req.df_top.to_dict()
        dic = dic['Value']

        if 'dates' in dic:
            res = dic['dates'].replace("'", '"')
            return json.loads(res)

        return Util.get_unique_list(df_res['date'])

    def unpack_rows_risk(self, ra):
        tmp = []

        for r in ra:
            tmp += r

        return tmp

    def _build_df_res_req(self):
        """
        """
        ra = False
        # df_res (response)
        yld_sh = []
        df_res_ra = pd.DataFrame()
        premiumspot_sh = []
        premiumfwd_sh = []
        vol_sh = []
        df_leg = self.obj_req.df_leg
        li_data = [f for e in self.raw_data for f in e]
        for e in li_data:
            if 'greeks' in e:
                for greek in ['delta', 'gamma', 'vega', 'theta']:
                    e[greek] = e['greeks'][greek]
                e.pop('greeks')
            if 'riskAnalysis' in e:
                ra = True
                for r in e['riskAnalysis']:
                    if r['name'] == "yieldShock":
                        yld_sh += self.unpack_rows_risk(r['rows'])
                    elif r['name'] == "premiumSpotShock":
                        premiumspot_sh += self.unpack_rows_risk(r['rows'])
                    elif r['name'] == "premiumForwardShock":
                        premiumfwd_sh += self.unpack_rows_risk(r['rows'])
                    elif r['name'] == "volNormalShock":
                        vol_sh += self.unpack_rows_risk(r['rows'])

                e.pop('riskAnalysis')

        df_res = pd.DataFrame(li_data)


        if ra is True:
            N_dates_ra = len(self.obj_req.riskAnalysis['dates'])
            N_fwd_ra = len(self.obj_req.riskAnalysis['forwards'])
            df_res_ra["yieldShock"] = yld_sh
            df_res_ra["premiumSpotShock"] = premiumspot_sh
            df_res_ra["premiumForwardShock"] = premiumfwd_sh
            df_res_ra["volNormalShock"] = vol_sh
            dt_ = []

            for d in df_res['date']:
                dt_ += [d]*N_dates_ra*N_fwd_ra

            df_res_ra['date'] = dt_

        # build list of dates
        li_date = self._get_dates(df_res)
        N = len(li_date)

        # df_req (request)
        # the order of results is by order of input
        # for each input the order of dates - but this is changed below
        # duplicate df_leg by number of dates

        df_leg = self.obj_req.df_leg

        df_req = pd.concat([df_leg] * N,
                           axis=0).reset_index(drop=True)
        if ra is True:
            df_req_ra = pd.DataFrame()

            for i in range(len(df_leg)):

                tmp = pd.concat([pd.DataFrame([df_leg.loc[i]])] * N_dates_ra*N_fwd_ra*N,
                                axis=0).reset_index(drop=True)

                tmp['shock_forwards'] = self.obj_req.riskAnalysis['forwards']*N_dates_ra*N
                dt_ = []
                for d in self.obj_req.riskAnalysis['dates']:
                    dt_ += [d]*N_fwd_ra

                tmp['shock_dates'] = dt_*N
                df_req_ra = pd.concat([df_req_ra, tmp], axis=0)
            df_req_ra = df_req_ra.reset_index(drop=True)

            df_req_ra['date'] = df_res_ra['date']

        # reorder results by date then initial order (tag)
        df_res['tag'] = range(len(df_res))
        df_res = df_res.sort_values(['date', 'tag']).reset_index(drop=True)

        #df_res_ra['tag']= range(len(df_res_ra))
        #df_res_ra= df_res_ra.sort_values(['date', 'tag']).reset_index(drop=True)
        # move date from df_res to df_req (more natural)
        # print(df_res['date'])
        df_req['date'] = pd.to_datetime(df_res['date'].copy())

        if ra is True:
            df_req_ra['date'] = pd.to_datetime(df_res_ra['date'].copy())
            df_res_ra = df_res_ra.drop('date', axis=1)

        df_res = df_res.drop('date', axis=1)

        df_res = df_res.rename(columns={
            'strike': 'strike_res',
            'nominal': 'nominal_res',
        })

        if 'error' not in df_res:
            df_res['error'] = 'No error'
        else:
            df_res['error'] = df_res['error'].fillna('No error')

        # move col error to last position
        cols = [c for c in df_res.columns if c != 'error']+['error']
        df_res = df_res[cols]

        # replace NaN returned by API
        df_res = df_res.replace('NaN', np.nan)

        # join df_req and df_res to make df_set
        df_set = pd.concat([df_req, df_res], axis=1)

        if ra is True:
            df_set_ra = pd.concat([df_req_ra, df_res_ra], axis=1)
            return df_req, df_res, df_set, df_req_ra, df_res_ra, df_set_ra
        else:
            return df_req, df_res, df_set, None, None, None

    def _build_dic_param(self):
        """
        """
        dic_req = self.df_req.to_dict()
        dic_req_param = {k: Util.get_unique_list(v.values())
                         for k, v in dic_req.items()}

        dic_data = self.df_res.to_dict()
        dic_res_param = {k: Util.get_unique_list(v.values())
                         for k, v in dic_data.items()}

        return dic_req_param, dic_res_param

    def save(self,
             folder_save='dump',
             name=None,
             tagged=True,
             excel=False):
        """
        """
        if name is None:
            name = 'SG_Research_ROTB'

        save_result(self.df_set,
                    folder_save, name=name + '_Components_response',
                    tagged=tagged,
                    excel=excel)

    def _repr_html_(self):
        """
        """
        return self.df_res.to_html()

    def info(self):
        """
        """
        md = """
A PostprocessROTB object from ComputeStrategyComponents endpoint has the properties:
+ `df_req`: request data (dataframe)
+ `df_res`: response data (dataframe)
+ `df_set`: request and response data combined (dataframe)

+ `df_req_ra`: request data for risk analysis (dataframe), None otherwise
+ `df_res_ra`: response data for risk analysis (dataframe), None otherwise
+ `df_set_ra`: request and response data combined for risk analysis (dataframe), None otherwise

+ `dic_req_param`: params in request, each param contains a list of all values taken (dictionary)
+ `dic_res_param`: params in response, each param contains a list of all values taken (dictionary)

+ `raw_data`: raw data in response under key 'componentSeries' (dictionary)

and the methods:
+ `save()` to save the data as `.csv` and `.xlsx` files
        """
        return Markdown(md)




import os

import datetime as dt
import pandas as pd

from IPython.display import Markdown

from ._util import Util


class ResponseRotbCurves:
    """
    """

    def __init__(self,
                 dic_res=None):
        """
        """
        assert isinstance(dic_res, dict), \
            'Error: dic_res must be a dict'
        assert 'curves' in dic_res, \
            'Error: curves must be a key of dic_res'

        self.dic_res = dic_res
        self.preview = Util.preview(dic_res)
        self.df_res = self._build_df_res()

    def _build_df_res(self):
        """
        """
        return pd.DataFrame(self.dic_res['curves'])

    def save_result(self,
                    folder_save='dump',
                    name=None,
                    tagged=True,
                    excel=False):
        """
        """
        if not os.path.exists(folder_save):
            os.makedirs(folder_save)
        if name is None:
            name = self.__class__
        tag = ''
        if tagged:
            tag = dt.datetime.now().strftime('_%Y%m%d_%H%M%S')
        suffix = '.csv'
        filename = '{}{}{}'.format(name, tag, suffix)
        path = os.path.join(folder_save, filename)
        self.df_res.to_csv(path, index=None)
        print('file {} saved'.format(path))
        if excel:
            suffix = '.xlsx'
            filename = '{}{}{}'.format(name, tag, suffix)
            path = os.path.join(folder_save, filename)
            self.df_res.to_excel(path, index=None)
            print('file {} saved'.format(path))

    def _repr_html_(self):
        """
        """
        return self.df_res.to_html()

    def info(self):
        """
        """
        md = """
A ResponseRotbCurves object has the properties:
+ `dic_res`: raw response (dict)
+ `preview`: preview of raw response (string)
+ `df_res`: response data (dataframe)


and the methods:
+ `save_result()` to save the data as `.csv` and `.xlsx` files
        """
        return Markdown(md)



# -*- coding: utf-8 -*-
"""
Created on Thu Jun  7 12:38:22 2018

@author: kferret061312
"""
import itertools
from IPython.display import display


class riskAnalysis:
    def __init__(self):
        self.dates = None
        self.forwards = None
        self.attr = ['dates', 'forwards']
        self.cartesian = False

    def make_list(self):
        att_l = self.attr.copy()
        for attr in att_l:
            att = getattr(self, attr)
            if att is not None:
                setattr(self, attr, att.split(','))
            else:
                self.attr.remove(attr)

    def longest_(self):
        assert 'dates' in self.attr, """ User has to set up at least one date """

        assert len(self.attr) == 2, """ User had to set up at least one forward shock"""
        ll = 0
        for attr in self.attr:
            l = len(getattr(self, attr))
            if l > ll:
                max_attr = attr
            ll = l
        self.longest = max_attr

    def cartesian_(self):

        longest = getattr(self, self.longest)
        attr_l = self.attr.copy()
        attr_l.remove(self.longest)
        if len(attr_l) > 2:
            prod_short = list(itertools.product(getattr(self, attr_l[0]), getattr(self, attr_l[1])))

            short_1 = [n[0] for n in prod_short]
            short_2 = [n[1] for n in prod_short]
            prod_1 = list(itertools.product(longest, short_1))
            prod_2 = list(itertools.product(longest, short_2))

            setattr(self, self.longest, [p[0] for p in prod_1])
            setattr(self, attr_l[0], [p[1] for p in prod_1])
            setattr(self, attr_l[1], [p[1] for p in prod_2])
        else:
            prod_1 = list(itertools.product(longest, getattr(self, attr_l[0])))
            setattr(self, self.longest, [p[0] for p in prod_1])
            setattr(self, attr_l[0], [p[1] for p in prod_1])

    def expand(self, verbose=False):
        self.make_list()

        if self.cartesian is True:
            self.longest_()
            self.cartesian_()
        else:
            assert self.forwards is not None, """ User has to set at least one forward shock"""
            # for i in self.attr[1:]:
            # assert len(getattr(self,self.attr[0])) == len(getattr(self,i)), """If
            # not a cartesian product then length of {} should be equal to length of {}""".format(self.attr[0],self.attr[1])

        self.riskAnalysis = {attr: getattr(self, attr) for attr in self.attr}

        if verbose is True:
            display(self.riskAnalysis)



import numpy as np
import pandas as pd
import ezhc as hc
from scipy import interpolate

ROUND_NB_DIGIT = 2


def vol_unit(unit):
    """
    to help express vol either in bp/day or bp/year
    """
    u = unit.upper()
    if u == 'DAY' or u == 'D':
        return 1 / 252 ** 0.5
    else:
        return 1


def get_diff(data, start_date, end_date):
    """
    get difference between 2 dates
    """
    mask_1 = (data.index.get_level_values('date') == end_date)
    mask_2 = (data.index.get_level_values('date') == start_date)
    one = data.loc[mask_1]
    one.index = one.index.droplevel('date')
    two = data.loc[mask_2]
    two.index = two.index.droplevel('date')
    # res=res.index.droplevel('date')
    return one - two


def get_max(data, resize=False, *args):
    """
    get max
    TBC
    """
    if resize:
        data = select_data(data, *args)
    x = data.groupby(level='expiry').apply(max)
    return x.reindex(data.index.get_level_values('expiry').unique(), axis=0)


def get_min(data, resize=False, *args):
    """
    get min
    TBC
    """
    if resize:
        data = select_data(data, *args)
    x = data.groupby(level='expiry').apply(min)
    return x.reindex(data.index.get_level_values('expiry').unique(), axis=0)


def get_surface(data, date):
    """
    get surface for given date
    TBC
    """
    mask = (data.index.get_level_values('date') == date)
    x = data.loc[mask]
    return drop_level(x, 'date')


def select_data(data, args):
    """
    select data from dataframe
    args:
        for discrete values
            field, value1, value2, etc
        for range of values
            field, value_start, value_end, 'range'

    returns subset dataframe
    TBC
    """
    arg = args
    if arg[0] == 'date':
        if len(arg) == 4 and arg[3] == 'range':
            mask = (data.index.get_level_values(arg[0]) >= arg[1]) & (
                data.index.get_level_values(arg[0]) <= arg[2])
        else:
            mask = (data.index.get_level_values(arg[0]) == arg[1])
            for ar in arg[2:len(arg)]:
                mask = mask | (data.index.get_level_values(arg[0]) == ar)
    else:
        mask = (data.index.get_level_values(arg[0]) == arg[1])
        for ar in arg[2:len(arg)]:
            mask = mask | (data.index.get_level_values(arg[0]) == ar)

    return data.loc[mask]


def drop_level(data, index):
    """
    drop level of input dataframe - inplace
    TBC
    """
    data.index = data.index.droplevel(index)
    return data


def get_percentile(data, date, resize=False, *args):
    """
    get percentile rank from input dataframe
    TBC
    """
    if resize:
        data = select_data(data, *args)
    return get_surface(data.groupby(level='expiry').rank(pct=True), date)


def get_mean(data, resize=False, *args):
    """
    get mean
    TBC
    """
    if resize:
        data = select_data(data, *args)
    data = data.dropna()
    x = data.groupby(level='expiry').apply(np.mean)
    return x.reindex(data.index.get_level_values('expiry').unique(), axis=0)


def get_std(data, resize=False, *args):
    """
    get standard deviation
    TBC
    """
    if resize:
        data = select_data(data, *args)
    data = data.dropna()
    x = data.groupby(level='expiry').apply(np.std)
    return x.reindex(data.index.get_level_values('expiry').unique(), axis=0)


def get_z_score(data, end_dt, resize=False, *args):
    """
    get z-score
    TBC
    """
    if resize:
        data = select_data(data, *args)
    return (get_surface(data, end_dt) - get_mean(data)) / get_std(data)


def build_weight(spread, weight, **kwargs):
    """
    build weight
    TBC
    """
    if (len(spread) is not len(weight)):
        if kwargs == {}:
            kwargs.setdefault('type', 'S')
        else:
            kwargs = kwargs['kwargs']

        if (len(spread) % 2 == 0 or (len(spread) % 2 == 0 and len(spread) % 3 == 0)) and (
                kwargs['type'].upper() != 'FLY' or kwargs['type'].upper() != 'F'):
            print('spread - weight len mismatch by default spread weighted (-1,1)')
            w = np.repeat(1, len(spread))
            pos = np.arange(1, len(w), 2)
            w[pos] = w[pos] * -1
        elif (len(spread) % 3 == 0 or (len(spread) % 2 == 0 and len(spread) % 3 == 0)) and (
                kwargs['type'].upper() == 'FLY' or kwargs['type'].upper() == 'F'):
            print('fly - weight len mismatch by default fly weighted (-1,2,-1)')
            w = np.repeat(-1, len(spread))
            pos = np.arange(1, len(w), 2)
            w[pos] = w[pos] * -2
    else:
        w = weight
    return w


# def build_spread(data, spread, typ_, resize=False, *args):
#     """
#     build spread
#     TBC
#     """
#     if resize:
#         data = select_data(data, args)
#     res = pd.DataFrame()
#     if typ_.upper() == "S" or typ_.upper() == "SLOPE":
#         for s in spread:
#             res[s[0] + s[1]] = data[s[1]] - data[s[0]]
#         return res
#     if typ_.upper() == "C" or typ._upper() == "CALENDAR":
#         from collections import OrderedDict
#         dic = OrderedDict()
#         for s in spread:
#             dic[(s[0] + s[1])] = data.loc[s[1]] - data.loc[s[0]]
#         res = pd.concat(dic.values(), keys=dic.keys())
#         res.index.names = ('expiry', 'date')
#         return res


def build_spread(data, spread, axis="T", weight=[], resize=False, *args, **kwargs):
    """
    build weight
    TBC
    """
    if resize:
        data = select_data(data, args)
    res = pd.DataFrame()
    w = build_weight(spread[0], weight, **kwargs)
    if axis.upper() == 'T' or axis.upper() == 'TENOR':
        for s in spread:
            nme = str()
            for i, e in enumerate(s):
                if i == 0:
                    res_tmp = data[s[i]] * w[i]
                else:
                    res_tmp = res_tmp + data[s[i]] * w[i]
                nme = nme + s[i]
            res[nme] = res_tmp
        return res

    if axis.upper() == 'E' or axis.upper() == 'EXPIRY':
        from collections import OrderedDict
        dic = OrderedDict()
        idx = data.loc[spread[0][0]].index
        col = data.loc[spread[0][0]].columns
        for s in spread:
            res_tmp = pd.DataFrame(0, index=idx, columns=col)
            nme = str()
            for i, e in enumerate(s):
                res_tmp = res_tmp + data.loc[s[i]] * w[i]
                nme = nme + s[i]
            dic[nme] = res_tmp
        res = pd.concat(dic.values(), keys=dic.keys())
        res.index.names = ('expiry', 'date')
        return res


def select(vol, expiries, tenors):
    if 'tenor' in vol.columns.names and 'expiry' in vol.index.names:
        return vol[tenors].loc[expiries]
    elif 'tenor' in vol.index.names and 'expiry' in vol.columns.names:
        return vol[expiries].loc[tenor]
    elif 'tenor' in vol.columns.names and 'expiry' in vol.columns.names:
        if vol.columns.names[0] is 'tenor':
            return vol.loc[:, (tenors, expiries)]
        else:
            return vol.loc[:, (expiries, tenors)]
    elif 'tenor' in vol.index.names and 'expiry' in vol.index.names:
        if vol.index.names[0] is 'tenor':
            return vol.loc[(tenors, expiries), :]
        else:
            return vol.loc[(expiries, tenors), :]


def generate_couples(start, end, step):
    return [[round(start + i * step, 2), round(end - i * step, 2)] for i in range(int((0.5 - start) / step))]


def volatility_cone(vol, expiries, tenors, start_date, start=0, end=1, step=0.25, plot=True):
    exp = expiries.split(',')
    ten = tenors.split(',')
    assert len(exp) is 1 or len(ten) is 1, """ one of expiries or tenors should take a single value """
    df = select(vol, exp, ten)
    df2 = df.loc[start_date:].quantile(q=np.arange(start, end + step, step))
    df2.loc['last'] = df.iloc[-1]
    return df2


def ewma(data, window=None, alpha=None):
    if alpha is None:
        alpha = 2 / (window + 1.0)
        alpha_rev = 1 - alpha
    else:
        alpha_rev = alpha
        alpha = 1 - alpha_rev

    n = data.shape[0]

    pows = alpha_rev ** (np.arange(n + 1))
    scale_arr = 1 / pows[:-1]
    offset = data[0] * pows[1:]
    pw0 = alpha * alpha_rev ** (n - 1)

    mult = data * pw0 * scale_arr
    cumsums = mult.cumsum()
    out = offset + cumsums * scale_arr[::-1]
    return out


def built_quantile_series(df, list_of_quantiles, colors):
    df = df.T
    df.columns = [round(float(d), 2) if d is not 'last' else d for d in df.columns]

    quantiles = []
    k = 0
    for i in list_of_quantiles:
        quantiles.append({
            'name': str(i[0] * 100) + ' ' + str(i[1] * 100) + ' percentile.',
            'data': [[list(df[i[0]].to_dict().values())[j], list(df[i[1]].to_dict().values())[j]] for j in
                     range(len(df.index))],
            'type': 'arearange',
            'color': colors[k],
            'marker': {
                'enabled': 'false'
            }
        })
        k += 1
    res = quantiles + [{'name': 'median',
                        'data': list(df[0.5].to_dict().values()),
                        'color': '#EEEEEE',
                        'lineWidth': 5,
                        'marker': {
                            'enabled': 'false'
                        }}, ] + [{'name': 'last',
                                  'data': list(df['last'].to_dict().values()),
                                  'color': '#FF0000',
                                  'lineWidth': 5,
                                  'marker': {'symbol': "circle",
                                             'radius': 5}}, ]
    return res


def vol_cone_plot(series, iv_rea, exp_ten, dt, xaxis=None, unit=' (bp/d)', save=False, save_path=None):
    g = hc.Highcharts()
    g.chart.height = 550
    g.legend.enabled = True
    g.legend.layout = 'horizontal'
    g.legend.align = 'center'
    g.legend.maxHeight = 100
    # g.tooltip.crosshairs = True
    g.tooltip.shared = True
    g.tooltip.valueSuffix = unit
    g.tooltip.valueDecimals = 2
    g.chart.zoomType = 'x'
    g.title.text = iv_rea + ' Volatility cone on ' + exp_ten + ' from ' + dt

    g.yAxis = [{'title': {'text': unit}},
               {'title': {'text': unit}, 'opposite': True}]

    g.xAxis = [{
        'categories': xaxis  # ['1M10Y', '3M10Y', '6M10Y', '1Y10Y','2Y10Y','5Y10Y']
    }]

    g.series = series + []
    display(g.plot(save=save))


def generate_relative_strikes(start, end, step):
    l = np.arange(start, end + step, step)
    return ",".join(['@' + str(ll) + 'bp' for ll in l])


def local_interpolation_one_by_two(df, idx, nb_of_points=2, step=0.001):
    c = list(df.columns)
    index = df.index
    width = []
    error = []
    k = 0
    for i in idx:
        ind = index[k]
        idx_ = c.index(i)
        min_ = c[idx_ - nb_of_points]
        max_ = c[idx_ + nb_of_points]
        data = df.loc[ind, min_:max_]
        strikes = [d.replace('@', '') for d in data.index]
        strikes = [float(s.replace('bp', '')) for s in strikes]
        tck = interpolate.splrep(strikes, data, s=0)
        xnew = np.arange(strikes[0], strikes[-1] + step, step)
        ynew = interpolate.splev(xnew, tck, der=0)
        ynew = pd.DataFrame([ynew], columns=xnew)
        ynew_abs = ynew.abs()
        width.append(float(ynew_abs.idxmin(axis=1)) / 2)
        ii = list(xnew).index(float(ynew_abs.idxmin(axis=1)))
        error.append(ynew.iloc[0, ii])
        k += 1
    return pd.concat([pd.DataFrame(width, columns=['width'], index=index).abs(),
                      pd.DataFrame(error, columns=['error'], index=index)], axis=1)



# -*- coding: utf-8 -*-
"""
Created on Wed Jun 13 10:59:02 2018

@author: kferret061312
"""


import os
import copy
import functools

import datetime as dt
import pandas as pd

from IPython.display import Markdown
from dateutil.parser import parse

from ._util import Util
from .response_rotb_compute_strategy_prices import ResponseRotbComputeStrategyPrices
from sgmarkets_api_auth.util import save_result


class SliceRotbComputeStrategyPricesRiskAnalysis:
    """
    """

    def __init__(self,
                 obj_res,
                 x=None,
                 y=None,
                 z=None,
                 dic_req_fix=None,
                 value=None,
                 y_pos='index',  # index or column
                 ):
        """
        """
        # print(type(obj_res))
        assert isinstance(obj_res, ResponseRotbComputeStrategyPrices), \
            'Error: obj_res must be a ResponseRotbComputeStrategyPrices object'
        self.res = obj_res
        self.y_pos = y_pos

        msg1 = 'Error: {} must be a column name in df_res_ra'
        msg1b = 'Error: y_pos must be a "index" or "column"'
        msg2 = 'z must be None if y is None'
        msg3 = 'Error: dic_req_fix must be a dict'
        msg4 = 'Error: {} is not a column of df_req_ra'
        msg5 = 'Error: {} has {} values - Must be one'
        msg6 = 'Error: value must be a column of df_res_ra'
        msg7 = 'Error: {} is present in dic_req_fix'
        msg8 = 'Error: {} has {} values in dic_res_param - Must be one'

        assert x in self.res.df_res_ra.columns, msg1.format('x')
        if y:
            assert y in self.res.df_res_ra.columns, msg1.format('y')
            assert y_pos in ['index', 'column'], msg1b
        if x and y:
            if z is not None:
                assert z in self.res.df_res_ra.columns, msg1.format('z')
        else:
            assert z is None, msg2

        assert isinstance(dic_req_fix, dict), msg3
        assert value in self.res.df_res_ra.columns, msg4.format(value)

        li_xyz = [e for e in [x, y, z] if e is not None]
        if len(li_xyz) == 1:

            df_slice = obj_res.df_res_ra[[x, value]]
            df_pivot = df_slice.set_index([x], inplace=True)

        if len(li_xyz) == 2:
            # display(df_slice)
            df_pivot = obj_res.df_res_ra.pivot_table(index=x,
                                                     columns=y,
                                                     values=value)
            # display(df_pivot)

        if len(li_xyz) == 3:
            df_pivot = obj_res.df_res_ra[[x, y, z, value]]

            if self.y_pos == 'index':
                df_pivot = df_pivot.pivot_table(index=[x, y],
                                                columns=z,
                                                values=value)

                df_pivot = df_pivot.reindex(obj_res.df_res_ra[x].drop_duplicates(),
                                            axis=0,
                                            level=0)
                df_pivot = df_pivot.reindex(obj_res.df_res_ra[y].drop_duplicates(),
                                            axis=0,
                                            level=1)

                df_pivot = df_pivot.reindex(obj_res.df_res_ra[z].drop_duplicates(),
                                            axis=1)

            if self.y_pos == 'column':
                df_pivot = df_pivot.pivot_table(index=x,
                                                columns=[y, z],
                                                values=value)
                df_pivot = df_pivot.reindex(obj_res.df_res_ra[x].drop_duplicates(),
                                            axis=0)
                df_pivot = df_pivot.reindex(obj_res.df_res_ra[y].drop_duplicates(),
                                            axis=1,
                                            level=0)
                df_pivot = df_pivot.reindex(obj_res.df_res_ra[z].drop_duplicates(),
                                            axis=1,
                                            level=1)

        self.df_pivot = df_pivot
        # transform

    def save(self,
             folder_save='dump',
             name=None,
             tagged=True,
             excel=False):
        """
        """
        if name is None:
            name = 'SG_Research_ROTB'

        save_result(self.df_pivot,
                    folder_save, name=name + '_Prices_RA_response',
                    tagged=tagged,
                    excel=excel)

    def _repr_html_(self):
        """
        """
        return self.df_slice_ra.to_html()

    def info(self):
        """
        """
        md = """
A SliceRotbComputeStrategyComponents object has the following properties:
+ `df_slice`: dataframe extracted from the postprocess dataframe `df_set` so that:
    + No constraints on params `x`, `y` (if input), `z` (if input)
    + All other request params are kept only if their value match the `dic_req_fix` dictionary of fixed values
        + All request params not `x` or `y` or `z` having more than one value must be specified in `dic_req_fix`
    + Only the response param `value` is kept

+ `df_pivot`: dataframe extracted from `df_slice`. It can be of dimension:
+ 1 if only `x` is defined
+ 2 if `x`and `y` are defined
+ 3 if `x`and `y` and `z` are defined
    + If the dimension is 3 and `y_pos` is `index` then (`x`, `y`) are a dataframe row multi-index
    + If the dimension is 3 and `y_pos` is `column` then (`y`, `z`) are a dataframe column multi-index
        """
        return Markdown(md)




from ._obj_from_dict import ObjFromDict

from .request_rotb_curves import RequestRotbCurves
from .response_rotb_curves import ResponseRotbCurves

from .request_rotb_compute_strategy_components import RequestRotbComputeStrategyComponents
from .response_rotb_compute_strategy_components import ResponseRotbComputeStrategyComponents
from .slice_rotb_compute_strategy_components import SliceRotbComputeStrategyComponents
from .slice_rotb_compute_strategy_components_risk_analysis import SliceRotbComputeStrategyComponentsRiskAnalysis

from .request_rotb_compute_strategy import RequestRotbComputeStrategy
from .response_rotb_compute_strategy import ResponseRotbComputeStrategy
from .slice_rotb_compute_strategy import SliceRotbComputeStrategy
from .slice_rotb_compute_strategy_risk_analysis import SliceRotbComputeStrategyRiskAnalysis

from .request_rotb_compute_strategy_prices import RequestRotbComputeStrategyPrices
from .response_rotb_compute_strategy_prices import ResponseRotbComputeStrategyPrices
from .slice_rotb_compute_strategy_prices import SliceRotbComputeStrategyPrices
from .slice_rotb_compute_strategy_prices_risk_analysis import SliceRotbComputeStrategyPricesRiskAnalysis

dic_endpoint = {
    'v1_curves': {
        'request': RequestRotbCurves,
        'response': ResponseRotbCurves,
    },
    'v1_compute_strategy_components': {
        'request': RequestRotbComputeStrategyComponents,
        'response': ResponseRotbComputeStrategyComponents,
        'slice': SliceRotbComputeStrategyComponents,
        'slice_riskAnalysis': SliceRotbComputeStrategyComponentsRiskAnalysis
    },
    'v1_compute_strategy': {
        'request': RequestRotbComputeStrategy,
        'response': ResponseRotbComputeStrategy,
        'slice': SliceRotbComputeStrategy,
        'slice_riskAnalysis': SliceRotbComputeStrategyRiskAnalysis,
    },
    'v1_compute_strategy_prices': {
        'request': RequestRotbComputeStrategyPrices,
        'response': ResponseRotbComputeStrategyPrices,
        'slice': SliceRotbComputeStrategyPrices,
        'slice_riskAnalysis': SliceRotbComputeStrategyPricesRiskAnalysis,
    },
    # to add new endpoint here after creating the corresponding
    # Request Response and optionally Slice objects
}

endpoint = ObjFromDict(dic_endpoint)




import datetime as dt
import pandas as pd

from pandas.tseries.offsets import CustomBusinessDay
from pandas.tseries.holiday import USFederalHolidayCalendar


class Calendar:
    """
    """

    def __init__(self):
        """
        """
        self.HOLIDAY_CALENDAR = USFederalHolidayCalendar()
        self.BDAY_US = CustomBusinessDay(calendar=self.HOLIDAY_CALENDAR)

    def fmt_date(self, ts):
        """
        """
        if isinstance(ts, str):
            ts = pd.Timestamp(ts)
        if isinstance(ts, dt.datetime):
            ts = pd.to_datetime(ts)
        return ts

    def get_BD_fallback_next(self, ts):
        """
        """
        ts = self.fmt_date(ts)
        return ts + 0 * self.BDAY_US

    def get_BD_fallback_previous(self, ts):
        """
        """
        ts = self.fmt_date(ts)
        return ts + 1 * self.BDAY_US - 1 * self.BDAY_US


calendar = Calendar()




import statsmodels.api as sm
from statsmodels.sandbox.regression.predstd import wls_prediction_std
import matplotlib.pyplot as plt
from sklearn import linear_model
from sklearn.metrics import mean_squared_error, r2_score
from sklearn.linear_model import HuberRegressor
from statsmodels.regression.quantile_regression import QuantReg
from statsmodels.tools.tools import add_constant
from IPython.display import display
import pandas as pd
import numpy as np
import ezhc as hc

plt.style.use('ggplot')

def hc_plot(title, series, x_name, y_name):
        g = hc.Highcharts()
        g.xAxis.title.text = x_name
        g.yAxis.title.text = y_name
        g.plotOptions = {
            'line': {
                'marker': {
                    'enabled': False
                }
            }
        }
        g.title = title
        g.series=series

        display(g.plot())

def OLS_stat_m(y, x, intercept=False, fit_test_ratio=0, summary=True, plot=True):
        
    split = int(round(len(x) * fit_test_ratio, 0))
    if split is 0:
        if intercept is True:
            x = sm.add_constant(x)
        model = sm.OLS(y, x)
        fit = model.fit()
        prstd, iv_l, iv_u = wls_prediction_std(fit)
        if summary:
            print(fit.summary())

        if plot:
            plt.figure(figsize=(12, 6))
            fig, ax = plt.subplots(figsize=(12, 6))
            ax.plot(x, y, 'o', label="data", color='black')
            ax.plot(x, fit.fittedvalues, 'r', label="OLS")
            ax.plot(x, iv_u, 'r')
            ax.plot(x, iv_l, 'r')
            ax.legend(loc='best')
        return fit, fit.params, fit.rsquared, prstd, iv_l, iv_u, pd.DataFrame(y - fit.fittedvalues, index=y.index)
    else:
        X_train = x.values[:split - 1]
        Y_train = y.values[:split - 1]
        X_test = x.values[split:]
        Y_test = y.values[split:]
        model = sm.OLS(Y_train, X_train)
        fit = model.fit()
        prstd, iv_l, iv_u = wls_prediction_std(fit)
        Y_pred = fit.predict(X_test)
        if summary:
            print(fit.summary())

        if plot:
            plt.figure(figsize=(12, 6))
            plt.scatter(X_train, Y_train, color='red')
            plt.scatter(X_test, Y_test, color='black')
            plt.plot(X_test, Y_pred, color='blue', linewidth=3)

    return fit, fit.params, fit.rsquared, prstd, iv_l, iv_u, pd.DataFrame(Y_test - Y_pred, index=list(y.index)[split:])


def OLS_scikit(Y, X, intercept=False, fit_test_ratio=0, plot=True):
    """
    Y: pd.Series
    X: pd.DataFrame
    """
    split = int(round(len(X)*fit_test_ratio, 0))
    xName = X.columns[0] if isinstance(X, pd.DataFrame) else X.name 
    yName = Y.columns[0] if isinstance(Y, pd.DataFrame) else Y.name

    if split is 0:
        X_train = X.values.reshape(len(X), 1)
        Y_train = Y.values.reshape(len(Y), 1)
        regr = linear_model.LinearRegression(fit_intercept=intercept)
        regr.fit(X_train, Y_train)
        Y_pred = regr.predict(X_train)
        if plot is True:
            data1 = [[X_train[i][0], Y_train[i][0]] for i in range(len(X_train))]
            data2 = [[X_train[i][0], Y_pred[i][0]] for i in range(len(X_train))]
            series = [
                {
                    'name':'Train',
                    'data':data1,
                    'type':'scatter'
                },
                {
                    'name':'OLS',
                    'data':data2,
                },
            ]
            hc_plot('OLS Scikit', series, xName, yName)
        return regr, regr.coef_,mean_squared_error(Y_train, Y_pred),r2_score(Y_train, Y_pred), pd.DataFrame(Y_train - Y_pred, index = Y.index)
    else:
        X_train=X.values[:split-1]
        Y_train=Y.values[:split-1]
        X_train = X_train.reshape(len(X_train), 1)
        Y_train = Y_train.reshape(len(X_train), 1)
        X_test = X.values[split:]
        Y_test = Y.values[split:]
        X_test = X_test.reshape(len(X_test), 1)
        Y_test = Y_test.reshape(len(Y_test), 1)
        regr = linear_model.LinearRegression(fit_intercept=intercept)
        regr.fit(X_train, Y_train)
        Y_pred = regr.predict(X_test)
        if plot is True:
            data1 = [[X_train[i][0], Y_train[i][0]] for i in range(len(X_train))]
            data2 = [[X_test[i][0], Y_test[i][0]] for i in range(len(X_test))]
            data3 = [[X_test[i][0], Y_pred[i][0]] for i in range(len(X_test))]
            series = [
                {
                    'name':'Train',
                    'data':data1,
                    'type':'scatter'
                },
                {
                    'name':'Test',
                    'data':data2,
                    'type':'scatter'
                },
                {
                    'name':'OLS',
                    'data':data3,
                },
            ]
            hc_plot('OLS Scikit', series, xName, yName)
        return regr, regr.coef_,mean_squared_error(Y_test, Y_pred),r2_score(Y_test, Y_pred), pd.DataFrame(Y_test - Y_pred, index=list(Y.index)[split:])


def RANSAC_scikit(Y, X, intercept=False, fit_test_ratio=0, plot=True):
    
    split = int(round(len(X) * fit_test_ratio, 0))
    xName = X.columns[0] if isinstance(X, pd.DataFrame) else X.name 
    yName = Y.columns[0] if isinstance(Y, pd.DataFrame) else Y.name

    if split is 0:
        X_train = X.values.reshape(len(X), 1)
        Y_train = Y.values.reshape(len(Y), 1)
        regr = linear_model.RANSACRegressor()
        regr.fit(X_train, Y_train)
        Y_pred = regr.predict(X_train)
        if plot is True:
            data1 = [[X_train[i][0], Y_train[i][0]] for i in range(len(X_train))]
            data2 = [[X_train[i][0], Y_pred[i][0]] for i in range(len(X_train))]
            x, y = X_train[regr.inlier_mask_], Y_train[regr.inlier_mask_]
            data3 = [[x[i][0], y[i][0]] for i in range(len(x))]
            series = [
                {
                    'name':'Outlier',
                    'data':data1,
                    'type':'scatter'
                },
                {
                    'name':'Pred',
                    'data':data2,
                },
                {
                    'name':'Train',
                    'data':data3,
                    'type':'scatter'
                },
            ]
            hc_plot('RANSAC Scikit',series, xName, yName)
            
            # plt.scatter(X_train, Y_train, color='yellow')
            # plt.scatter(X_train[regr.inlier_mask_], Y_train[regr.inlier_mask_], color='red')
            # plt.plot(X_train, Y_pred, color='blue', linewidth=3)
        return regr, regr.estimator_.coef_, mean_squared_error(Y_train, Y_pred), r2_score(Y_train,
                                                                                          Y_pred), pd.DataFrame(
            Y_train - Y_pred, index=Y.index), \
               regr.inlier_mask_
    else:
        X_train = X.values[:split - 1]
        Y_train = Y.values[:split - 1]
        X_train = X_train.reshape(len(X_train), 1)
        Y_train = Y_train.reshape(len(X_train), 1)
        X_test = X.values[split:]
        Y_test = Y.values[split:]
        X_test = X_test.reshape(len(X_test), 1)
        Y_test = Y_test.reshape(len(Y_test), 1)
        regr = linear_model.RANSACRegressor()
        regr.fit(X_train, Y_train)
        Y_pred = regr.predict(X_test)
        if plot is True:
            data1 = [[X_train[i][0], Y_train[i][0]] for i in range(len(X_train))]
            x, y = X_train[regr.inlier_mask_], Y_train[regr.inlier_mask_]
            data2 = [[x[i][0], y[i][0]] for i in range(len(x))]
            data3 = [[X_test[i][0], Y_test[i][0]] for i in range(len(X_test))]
            data4 = [[X_test[i][0], Y_pred[i][0]] for i in range(len(X_test))]
            
            series = [
                {
                    'name':'Outlier',
                    'data':data1,
                    'type':'scatter'
                },
                {
                    'name':'Train',
                    'data':data2,
                    'type':'scatter'
                },
                {
                    'name':'Test',
                    'data':data3,
                    'type':'scatter'
                },
                {
                    'name':'RANSAC',
                    'data':data4,
                }
            ]
            hc_plot('RANSAC Scikit',series, xName, yName)

            # plt.scatter(X_train, Y_train, color='yellow')
            # plt.scatter(X_train[regr.inlier_mask_], Y_train[regr.inlier_mask_], color='red')
            # plt.scatter(X_test, Y_test, color='black')
            # plt.plot(X_test, Y_pred, color='blue', linewidth=3)
        return regr, regr.estimator_.coef_, mean_squared_error(Y_test, Y_pred), r2_score(Y_test, Y_pred), pd.DataFrame(
            Y_test - Y_pred, index=list(Y.index)[split:]), \
               regr.inlier_mask_


def Huber_scikit(Y, X, intercept=False, fit_test_ratio=0, plot=True, alpha=0, max_iter=1000, eps=1.3):

    split = int(round(len(X) * fit_test_ratio, 0))
    xName = X.columns[0] if isinstance(X, pd.DataFrame) else X.name 
    yName = Y.columns[0] if isinstance(Y, pd.DataFrame) else Y.name
    if split is 0:
        X_train = X.values.reshape(len(X), 1)
        Y_train = Y.values
        regr = HuberRegressor(fit_intercept=intercept, alpha=alpha, max_iter=max_iter,
                              epsilon=eps)
        regr.fit(X_train, Y_train)
        Y_pred = regr.predict(X_train)
        if plot is True:
            x, y = X_train[regr.outliers_], Y_train[regr.outliers_]
            data1 = [[X_train[i][0], Y_train[i]] for i in range(len(X_train))]
            data2 = [[X_train[i][0], Y_pred[i]] for i in range(len(X_train))]
            data3 = [[x[i][0], y[i]] for i in range(len(x))]
            series = [
                {
                    'name':'Train',
                    'data':data1,
                    'type':'scatter'
                },
                {
                    'name':'Huber',
                    'data':data2,
                },
                {
                    'name':'Outlier',
                    'data':data3,
                    'type':'scatter'
                },
            ]
            hc_plot('Huber Scikit',series, xName, yName)
            
            # plt.scatter(X_train, Y_train, color='red')
            # plt.scatter(X_train[regr.outliers_], Y_train[regr.outliers_], color='yellow')
            # plt.plot(X_train, Y_pred, color='blue', linewidth=3)
        return regr, regr.coef_, mean_squared_error(Y_train, Y_pred), r2_score(Y_train, Y_pred), pd.DataFrame(
            Y_train - Y_pred, index=Y.index), \
               regr.outliers_
    else:
        X_train = X.values[:split - 1]
        Y_train = Y.values[:split - 1]
        X_train = X_train.reshape(len(X_train), 1)
        X_test = X.values[split:]
        Y_test = Y.values[split:]
        X_test = X_test.reshape(len(X_test), 1)
        regr = HuberRegressor(fit_intercept=intercept, alpha=0.0, max_iter=100,
                              epsilon=eps)
        regr.fit(X_train, Y_train)
        Y_pred = regr.predict(X_test)
        if plot is True:
            x, y = X_train[regr.outliers_], Y_train[regr.outliers_]
            data1 = [[X_train[i][0], Y_train[i]] for i in range(len(X_train))]
            data2 = [[x[i][0], y[i]] for i in range(len(x))]
            data3 = [[X_test[i][0], Y_test[i]] for i in range(len(X_test))]
            data4 = [[X_test[i][0], Y_pred[i]] for i in range(len(X_test))]
            
            series = [
                {
                    'name':'Train',
                    'data':data1,
                    'type':'scatter'
                },
                {
                    'name':'Outlier',
                    'data':data2,
                    'type':'scatter'
                },
                {
                    'name':'Test',
                    'data':data3,
                    'type':'scatter'
                },
                {
                    'name':'Huber',
                    'data':data4,
                }
            ]
            hc_plot('Huber Scikit',series, xName, yName)

            # plt.scatter(X_train, Y_train, color='red')
            # plt.scatter(X_train[regr.outliers_], Y_train[regr.outliers_], color='yellow')
            # plt.scatter(X_test, Y_test, color='black')
            # plt.plot(X_test, Y_pred, color='blue', linewidth=3)
        return regr, regr.coef_, mean_squared_error(Y_test, Y_pred), r2_score(Y_test, Y_pred), pd.DataFrame(
            Y_test - Y_pred, index=list(Y.index)[split:]), \
               regr.outliers_

        # http://www.gmelli.org/RKB/sklearn.linear_model.HuberRegressor


def TheilSen_scikit(Y, X, intercept=False, fit_test_ratio=0, plot=True):

    split = int(round(len(X) * fit_test_ratio, 0))
    xName = X.columns[0] if isinstance(X, pd.DataFrame) else X.name 
    yName = Y.columns[0] if isinstance(Y, pd.DataFrame) else Y.name

    if split is 0:
        X_train = X.values.reshape(len(X), 1)
        Y_train = Y.values
        regr = linear_model.TheilSenRegressor(fit_intercept=intercept)
        regr.fit(X_train, Y_train)
        Y_pred = regr.predict(X_train)
        Y_pred = [y[0] for y in Y_pred]
        if plot is True:
            data1 = [[X_train[i][0], Y_train[i]] for i in range(len(X_train))]
            data2 = [[X_train[i][0], Y_pred[i]] for i in range(len(X_train))]
            series = [
                {
                    'name':'Train',
                    'data':data1,
                    'type':'scatter'
                },
                {
                    'name':'TheilSen',
                    'data':data2,
                },
            ]
            hc_plot('Theilsen Scikit', series, xName, yName)

            # plt.scatter(X_train, Y_train, color='red')
            # # plt.scatter(X_train[regr.inlier_mask_], Y_train[regr.inlier_mask_],  color='red')
            # plt.plot(X_train, Y_pred, color='blue', linewidth=3)
        return regr, regr.coef_, mean_squared_error(Y_train, Y_pred), r2_score(Y_train, Y_pred), pd.DataFrame(
            Y_train - Y_pred, index=Y.index), \
               regr.breakdown_
    else:
        X_train = X.values[:split - 1]
        Y_train = Y.values[:split - 1]
        X_train = X_train.reshape(len(X_train), 1)
        X_test = X.values[split:]
        Y_test = Y.values[split:]
        X_test = X_test.reshape(len(X_test), 1)
        regr = linear_model.TheilSenRegressor(fit_intercept=intercept)
        regr.fit(X_train, Y_train)
        Y_pred = regr.predict(X_test)
        Y_pred = [y[0] for y in Y_pred]
        if plot is True:
            data1 = [[X_train[i][0], Y_train[i]] for i in range(len(X_train))]
            data2 = [[X_test[i][0], Y_test[i]] for i in range(len(X_test))]
            data3 = [[X_test[i][0], Y_pred[i]] for i in range(len(X_test))]
            series = [
                {
                    'name':'Train',
                    'data':data1,
                    'type':'scatter'
                },
                {
                    'name':'Test',
                    'data':data2,
                    'type':'scatter'
                },
                {
                    'name':'TheilSen',
                    'data':data3,
                },
            ]
            hc_plot('Theilsen Scikit', series, xName, yName)

            # plt.scatter(X_train, Y_train, color='red')
            # # plt.scatter(X_train[regr.inlier_mask_], Y_train[regr.inlier_mask_],  color='red')
            # plt.scatter(X_test, Y_test, color='black')
            # plt.plot(X_test, Y_pred, color='blue', linewidth=3)
        return regr, regr.coef_, mean_squared_error(Y_test, Y_pred), r2_score(Y_test, Y_pred), pd.DataFrame(
            Y_test - Y_pred, index=list(Y.index)[split:]), \
               regr.breakdown_


def quantile_regression_stat_m(Y, X, intercept=False, quantile=[0.5], fit_test_ratio=0, plot=True, summary=False):

    split = int(round(len(X) * fit_test_ratio, 0))

    if split is 0:
        if intercept is True:
            X_train = add_constant(X, prepend=True, has_constant='skip')
        else:
            X_train = X

        Y_train = Y
        regr = QuantReg(Y_train, X_train)
        R = {}
        Y_pred = {}
        for q in quantile:
            R[q] = regr.fit(q=q, vcov='robust', kernel='epa', bandwidth='hsheather', max_iter=1000, p_tol=1e-06)
            Y_pred[q] = R[q].predict(X_train)
            if summary is True:
                print(R[q].summary())
        if plot is True:
            plt.figure(figsize=(12, 6))
            for q in quantile:
                if q != 0.5:
                    plt.plot(X_train, Y_pred[q], linestyle='dotted', color='blue')
                else:
                    plt.plot(X_train, Y_pred[q], color='blue')

            plt.scatter(X_train, Y_train, color='red')
        return R, Y_pred
    else:
        if intercept is True:
            X = add_constant(X, prepend=True, has_constant='skip')
        X_train = X.values[:split - 1]
        Y_train = Y.values[:split - 1]
        X_test = X.values[split:]
        Y_test = Y.values[split:]
        regr = QuantReg(Y_train, X_train)
        R = {}
        Y_pred = {}
        for q in quantile:
            R[q] = regr.fit(q=q, vcov='robust', kernel='epa', bandwidth='hsheather', max_iter=1000, p_tol=1e-06)
            Y_pred[q] = R[q].predict(X_test)
            if summary is True:
                print(R[q].summary())
        if plot is True:
            plt.figure(figsize=(12, 6))
            for q in quantile:
                if q != 0.5:
                    plt.plot(X_test, Y_pred[q], linestyle='dotted', color='blue')
                else:
                    plt.plot(X_test, Y_pred[q], color='blue')

            plt.scatter(X_train, Y_train, color='red')
            plt.scatter(X_test, Y_test, color='black')
        return R, Y_pred




__name__ = 'sgmarkets_api_analytics_rotb'
name_url = __name__.replace('_', '-')

__version__ = '0.4.8'
__description__ = 'sgmarket api analytics ROTB helper'
__long_description__ = 'See repo README'
__author__ = 'sgmarkets'
__author_email__ = 'olivier.borderies@gmail.com'
__url__ = 'https://gitlab.com/{}/{}'.format(__author__,
                                            name_url)
__download_url__ = 'https://gitlab.com/{}/{}/repository/archive.tar.gz?ref={}'.format(__author__,
                                                                                      name_url,
                                                                                      __version__)
__keywords__ = ['sgmarkets', 'api', 'rotb']
__license__ = 'MIT'
__classifiers__ = ['Development Status :: 4 - Beta',
                   'License :: OSI Approved :: MIT License',
                   'Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6'
                   ]
__include_package_data__ = True
__package_data__ = {
    'markdown':
        ['markdown/info.md',
         ],
}
__zip_safe__ = False
__entry_points__ = {}




import os
from IPython.display import display, Markdown


def info():
    """
    """
    _dir = os.path.dirname(os.path.realpath(__file__))
    path = os.path.join(_dir, 'markdown', 'info.md')
    with open(path, 'r') as f:
        md = f.read()
        display(Markdown(md))



# -*- coding: utf-8 -*-
"""
Created on Mon Jun 18 15:30:04 2018

@author: kferret061312
"""
from .response_rotb_compute_strategy_components import ResponseRotbComputeStrategyComponents
from .response_rotb_compute_strategy_prices import ResponseRotbComputeStrategyPrices
from .slice_rotb_compute_strategy_prices_risk_analysis import SliceRotbComputeStrategyPricesRiskAnalysis
from .slice_rotb_compute_strategy_components_risk_analysis import SliceRotbComputeStrategyComponentsRiskAnalysis


class SliceRotbComputeStrategyRiskAnalysis:
    """
    """

    def __init__(self,
                 obj_res,
                 x=None,
                 y=None,
                 z=None,
                 dic_req_fix=None,
                 value=None,
                 y_pos='index',  # index or column
                 ):
        assert isinstance(obj_res, ResponseRotbComputeStrategyComponents) or isinstance(obj_res, ResponseRotbComputeStrategyPrices), \
            'Error: obj_res must be a ResponseRotbComputeStrategyComponents object or ResponseRotbComputeStrategyPrices'
        if isinstance(obj_res, ResponseRotbComputeStrategyComponents):
            sli = SliceRotbComputeStrategyComponentsRiskAnalysis(
                obj_res,
                x,
                y,
                z,
                dic_req_fix,
                value,
                y_pos,  # index or column
            )
            self.df_pivot = sli.df_pivot
        else:
            sli = SliceRotbComputeStrategyPricesRiskAnalysis(
                obj_res,
                x,
                y,
                z,
                dic_req_fix,
                value,
                y_pos,  # index or column
            )
            self.df_pivot = sli.df_pivot




import re
import pandas as pd

from collections import OrderedDict

from ._calendar import calendar as cal


class Util:

    @staticmethod
    def preview(data, start=200, end=200):
        """
        preview json returned by Api
        """
        string = str(data)
        n = len(string)

        if n <= start + end:
            return string

        beg, end = string[:start], string[-end:]
        res = beg + '.........({} more characters).........'.format(n) + end
        return res

    @staticmethod
    def get_unique_list(my_list):
        """
        """
        unique_list = list(OrderedDict((e, None) for e in my_list))
        unique_list = [e for e in unique_list if pd.notnull(e)]
        return unique_list

    @staticmethod
    def build_url(dic_url):
        """
        """
        return '{}{}{}'.format(dic_url['base_url'],
                               dic_url['service'],
                               dic_url['endpoint'])

    @staticmethod
    def get_list_bday_between(start_date, end_date):
        """
        """
        bd_range = pd.bdate_range(start_date,
                                  end_date,
                                  freq='C',
                                  holidays=cal.BDAY_US.holidays)
        return list(bd_range)

    @staticmethod
    def date_to_str(ts):
        """
        """
        return ts.strftime('%Y-%m-%d')

    @staticmethod
    def multiple_replace(dic, text):
        # Create a regex from the dict keys
        regex = re.compile('(%s)' % '|'.join(map(re.escape, dic.keys())))

        # For each match, lookup corresponding value in dict
        return regex.sub(lambda mo: dic[mo.string[mo.start():mo.end()]], text)




import os
import copy
import functools

import datetime as dt
import pandas as pd

from IPython.display import Markdown
from dateutil.parser import parse

from ._util import Util
from .response_rotb_compute_strategy_components import ResponseRotbComputeStrategyComponents
from sgmarkets_api_auth.util import save_result


class SliceRotbComputeStrategyComponents:
    """
    """

    def __init__(self,
                 obj_res,
                 x=None,
                 y=None,
                 z=None,
                 dic_req_fix=None,
                 value=None,
                 y_pos='index',  # index or column
                 ):
        """
        """
        # print(type(obj_res))
        assert isinstance(obj_res, ResponseRotbComputeStrategyComponents), \
            'Error: obj_res must be a ResponseRotbComputeStrategyComponents object'
        self.res = obj_res
        self.y_pos = y_pos

        self.res.dic_req_param = {k: list(self.res.df_req[k].drop_duplicates())
                                  for k in self.res.dic_req_param}

        msg1 = 'Error: {} must be a column name in df_req'
        msg1b = 'Error: y_pos must be a "index" or "column"'
        msg2 = 'z must be None if y is None'
        msg3 = 'Error: dic_req_fix must be a dict'
        msg4 = 'Error: {} is not a column of df_req'
        msg5 = 'Error: {} has {} values - Must be one'
        msg6 = 'Error: value must be a column of df_res'
        msg7 = 'Error: {} is present in dic_req_fix'
        msg8 = 'Error: {} has {} values in dic_res_param - Must be one'

        assert x in self.res.df_req.columns, msg1.format('x')
        if y:
            assert y in self.res.df_req.columns, msg1.format('y')
            assert y_pos in ['index', 'column'], msg1b
        if x and y:
            if z is not None:
                assert z in self.res.df_req.columns, msg1.format('z')
        else:
            assert z is None, msg2

        assert isinstance(dic_req_fix, dict), msg3

        # transform
        for k, v in dic_req_fix.items():
            if isinstance(v, pd.Timestamp):
                v = Util.date_to_str(v)
                dic_req_fix[k] = v
            if isinstance(v, (str, int, float)):
                dic_req_fix[k] = [v]

        for k, v in dic_req_fix.items():
            assert k in self.res.df_req.columns, msg4.format(k)
            assert len(v) == 1, msg5.format(k, len(v))

        assert value in self.res.df_res.columns, msg6

        li_req_fix = list(dic_req_fix.keys())
        li_req_sel = copy.deepcopy(li_req_fix)
        li_xyz = [e for e in [x, y, z] if e is not None]
        for v in li_xyz:
            li_req_sel.append(v)
            assert v not in li_req_fix, msg7.format(v)
        # display(li_xyz)

        dic_req_param_full = copy.deepcopy(dic_req_fix)
        # display(dic_req_fix)
        dic_req_param_full = {k: v[0] for k, v in dic_req_param_full.items()}
        # display(dic_req_param_full)

        for c in self.res.df_req.columns:
            if c not in li_req_sel:
                v = self.res.dic_req_param[c]
                assert len(v) == 1, \
                    msg8.format(c, len(self.res.dic_res_param))
                dic_req_param_full[c] = v[0]

        # display(dic_req_param_full)
        li_mask = []
        for k, v in dic_req_param_full.items():
            # EXCEPTION
            if k != 'nominal':
                li_mask.append(self.res.df_req[k] == v)
        # display(li_mask)
        mask = functools.reduce((lambda x, y: x & y),
                                li_mask)

        df_slice = pd.concat([self.res.df_req[li_xyz],
                              self.res.df_res[value]], axis=1)
        # display(df_slice)
        df_slice = df_slice.loc[mask]
        self.df_slice = df_slice.reset_index(drop=True)

        self.df_pivot = None

        if len(li_xyz) == 1:
            df_pivot = df_slice[[x, value]]

            df_pivot = df_pivot.set_index(x)
            df_pivot = df_pivot.loc[self.res.dic_req_param[x]]

        if len(li_xyz) == 2:
            # display(df_slice)
            df_pivot = df_slice.pivot_table(index=x,
                                            columns=y,
                                            values=value)
            # display(df_pivot)
            df_pivot = df_pivot.loc[self.res.dic_req_param[x],
                                    self.res.dic_req_param[y]]

        if len(li_xyz) == 3:
            df_pivot = df_slice[[x, y, z, value]]

            if self.y_pos == 'index':
                df_pivot = df_pivot.pivot_table(index=[x, y],
                                                columns=z,
                                                values=value)

                df_pivot = df_pivot.reindex(self.res.dic_req_param[x],
                                            axis=0,
                                            level=0)
                df_pivot = df_pivot.reindex(self.res.dic_req_param[y],
                                            axis=0,
                                            level=1)

                df_pivot = df_pivot.reindex(self.res.dic_req_param[z],
                                            axis=1)

            if self.y_pos == 'column':
                df_pivot = df_pivot.pivot_table(index=x,
                                                columns=[y, z],
                                                values=value)
                df_pivot = df_pivot.reindex(self.res.dic_req_param[x],
                                            axis=0)
                df_pivot = df_pivot.reindex(self.res.dic_req_param[y],
                                            axis=1,
                                            level=0)
                df_pivot = df_pivot.reindex(self.res.dic_req_param[z],
                                            axis=1,
                                            level=1)

        self.df_pivot = df_pivot

    def save(self,
             folder_save='dump',
             name=None,
             tagged=True,
             excel=False):
        """
        """
        if name is None:
            name = 'SG_Research_ROTB'

        save_result(self.df_pivot,
                    folder_save, name=name + '_Components_response',
                    tagged=tagged,
                    excel=excel)

    def _repr_html_(self):
        """
        """
        return self.df_slice.to_html()

    def info(self):
        """
        """
        md = """
A SliceRotbComputeStrategyComponents object has the following properties:
+ `df_slice`: dataframe extracted from the postprocess dataframe `df_set` so that:
    + No constraints on params `x`, `y` (if input), `z` (if input)
    + All other request params are kept only if their value match the `dic_req_fix` dictionary of fixed values
        + All request params not `x` or `y` or `z` having more than one value must be specified in `dic_req_fix`
    + Only the response param `value` is kept

+ `df_pivot`: dataframe extracted from `df_slice`. It can be of dimension:
+ 1 if only `x` is defined
+ 2 if `x`and `y` are defined
+ 3 if `x`and `y` and `z` are defined
    + If the dimension is 3 and `y_pos` is `index` then (`x`, `y`) are a dataframe row multi-index
    + If the dimension is 3 and `y_pos` is `column` then (`y`, `z`) are a dataframe column multi-index
        """
        return Markdown(md)
